package fitBut.fbPerceptionModule;

import fitBut.agents.FBAgent;
import fitBut.fbEnvironment.FBMap;
import fitBut.fbPerceptionModule.data.BlockType;
import fitBut.utils.logging.HorseRider;
import fitBut.fbPerceptionModule.data.ActionResult;
import fitBut.fbEnvironment.FBCells.objects.*;
import fitBut.fbEnvironment.utils.FBObjectType;
import fitBut.fbMultiagent.FBTask;
import fitBut.fbPerceptionModule.data.AgentInfo;
import fitBut.fbPerceptionModule.data.SimInfo;
import eis.iilang.Function;
import eis.iilang.Parameter;
import eis.iilang.ParameterList;
import eis.iilang.Percept;

import fitBut.utils.Point;

import java.util.*;

/**
 * @author : Vaclav Uhlir
 * @since : 12.9.2019
 **/
@SuppressWarnings("deprecation")
public class FBPerceptsProcessorV2 {

    private static final String TAG = "FBPerceptsProcessorV2";
    private SimInfo simInfo;
    private AgentInfo agentInfo;
    private Set<LinkedList<Parameter>> goals;
    private Set<LinkedList<Parameter>> obstacles;
    private Set<LinkedList<Parameter>> things;
    private Set<LinkedList<Parameter>> attached;
    private HashSet<Point> attachedIndex;

    public void processPercepts(Collection<Percept> percepts) {
        things = new HashSet<>();
        obstacles = new HashSet<>();
        goals = new HashSet<>();
        attached = new HashSet<>();
        simInfo.dumpTasks();

        for (Percept percept : percepts) {
            switch (percept.getName()) {
                //map objects
                case "thing":
                    things.add(percept.getParameters());
                    break;
                case "obstacle":
                    obstacles.add(percept.getParameters());
                    break;
                case "goal":
                    goals.add(percept.getParameters());
                    break;
                case "attached":
                    attached.add(percept.getParameters());
                    break;

                // Agent Information
                case "name":
                    agentInfo.setName(percept.getParameters().get(0).toString());
                    break;
                case "vision":
                    agentInfo.setVision(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "energy":
                    agentInfo.setEnergy(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "disabled":
                    agentInfo.setDisabled(percept.getParameters().get(0).toString().equals("true"));
                    break;
                case "team":
                    agentInfo.setTeam(percept.getParameters().get(0).toString());
                    break;

                // Sim information
                case "steps":
                    simInfo.setSteps(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "simStart":
                    simInfo.setSimStarted(true);
                    break;
                case "requestAction":
                    simInfo.setRequestingAction(true);
                    break;
                case "score":
                    simInfo.setScore(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "step":
                    simInfo.setStep(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "actionID":
                    simInfo.setActionID(Integer.parseInt(percept.getParameters().get(0).toString()));
                    break;
                case "lastAction":
                    simInfo.setLastAction(percept.getParameters().get(0).toString()); //TODO: parse to enum?
                    break;
                case "lastActionParams":
                    simInfo.setLastActionParams(percept.getParameters().get(0)); //TODO: parse to something useful?
                    break;
                case "deadline":
                    simInfo.setDeadline(Long.parseLong(percept.getParameters().get(0).toString()));
                    break;
                case "timestamp":
                    simInfo.setTimestamp(Long.parseLong(percept.getParameters().get(0).toString()));
                    break;
                case "lastActionResult":
                    simInfo.setLastActionResult(ActionResult.fromString(percept.getParameters().get(0).toString()));
                    break;
                case "task":
                    processTask(percept.getParameters());
                    break;

                default:
                    HorseRider.warn(TAG, "processPercepts: " + agentInfo.getName() + " " +
                            String.format("Skipped: %s - %s \n", percept.getName(), percept.getParameters()));

            }

        }
    }

    public void runMapUpdate(FBMap map, FBAgent agent) {
        Point pov = agent.getLocalPosition();
        map.insertEmptyVisionCells(agentInfo.getVision(), pov, simInfo.getStep());
        makeAttachedIndex();
        processThings(map, things, agent);
        processObstacles(map, obstacles, pov);
        processGoals(map, goals, pov);
        map.printMap();
    }

    private void makeAttachedIndex() {
        attachedIndex = new HashSet<>();
        for (LinkedList<Parameter> parameters : attached) {
            if (parameters.size() != 2) {
                HorseRider.yell(TAG, "makeAttachedIndex: " + agentInfo.getName() + " " +
                        "unknown format!: " + parameters);
            } else {
                attachedIndex.add(new Point(Integer.parseInt(parameters.get(0).toString()), Integer.parseInt(parameters.get(1).toString())));
            }
        }
    }

    private void processTask(LinkedList<Parameter> parameters) {
        if (parameters.size() != 4) {
            HorseRider.yell(TAG, "processTask: " + agentInfo.getName() + " " +
                    "unknown format!: " + parameters);
        } else {
            String taskName = parameters.get(0).toString();
            int taskDeadline = Integer.parseInt(parameters.get(1).toString());
            int taskReward = Integer.parseInt(parameters.get(2).toString());
            Set<LinkedList<Parameter>> taskStructure = new HashSet<>();
            for (Parameter block : ((ParameterList) parameters.get(3))) {
                taskStructure.add(((Function) block).getParameters());
            }
            simInfo.addTask(new FBTask(taskName, taskDeadline, taskReward, processTaskBlocks(taskStructure)));
        }
    }

    private HashMap<Point, BlockType> processTaskBlocks(Set<LinkedList<Parameter>> blocks) {
        HashMap<Point, BlockType> taskStructure = new HashMap<>();
        for (LinkedList<Parameter> parameters : blocks) {
            if (parameters.size() != 3) {
                HorseRider.yell(TAG, "processTaskBlocks: " + agentInfo.getName() + " " +
                        "unknown format!: " + parameters);
            } else {
                Point position = new Point(Integer.parseInt(parameters.get(0).toString()), Integer.parseInt(parameters.get(1).toString()));
                taskStructure.put(position, simInfo.getBlockType(parameters.get(2).toString()));
                //structureMap.addCellObject(position, new FBBlockObject(simInfo.getBlockType(parameters.get(2).toString()), true), simInfo.getStep());
            }
        }
        return taskStructure;
    }

    private void processGoals(FBMap map, Set<LinkedList<Parameter>> obstacles, Point pov) {
        for (LinkedList<Parameter> parameters : obstacles) {
            if (parameters.size() != 2) {
                HorseRider.yell(TAG, "processGoals: " + agentInfo.getName() + " " +
                        "unknown format!: " + parameters);
            } else {
                Point position = new Point(pov);
                //HorseRider.inquire(TAG, "processGoals: "+agentInfo.getName()+" goal at: "+
                //        parameters.get(0).toString()+"\t"+parameters.get(1).toString());
                position.translate(Integer.parseInt(parameters.get(0).toString()), Integer.parseInt(parameters.get(1).toString()));
                map.addCellObject(position, new FBOtherObject(FBObjectType.__FBGoal), simInfo.getStep());
            }

        }
    }

    private void processObstacles(FBMap map, Set<LinkedList<Parameter>> obstacles, Point pov) {
        for (LinkedList<Parameter> parameters : obstacles) {
            if (parameters.size() != 2) {
                HorseRider.yell(TAG, "processObstacles: " + agentInfo.getName() + " " +
                        "unknown format!: " + parameters);
            } else {
                Point position = new Point(pov);
                //HorseRider.inquire(TAG, "processObstacles: "+agentInfo.getName()+" obstacle at: "+
                //        parameters.get(0).toString()+"\t"+parameters.get(1).toString());
                position.translate(Integer.parseInt(parameters.get(0).toString()), Integer.parseInt(parameters.get(1).toString()));
                map.addCellObject(position, new FBOtherObject(FBObjectType.__FBObstacle), simInfo.getStep());
            }

        }
    }

    private void processThings(FBMap map, Set<LinkedList<Parameter>> things, FBAgent agent) {
        Point pov = agent.getLocalPosition();
        for (LinkedList<Parameter> parameters : things) {
            if (parameters.size() != 4) {
                HorseRider.yell(TAG, "processThings: " + agentInfo.getName() + " " +
                        "unknown format!: " + parameters);
            } else {
                Point position = new Point(pov); // get agent position
                Point relative = new Point(Integer.parseInt(parameters.get(0).toString()), Integer.parseInt(parameters.get(1).toString()));
                position.translate(relative); //translate
                switch (parameters.get(2).toString()) {
                    case "marker":
                        map.addCellObject(position, new FBMarkerObject(parameters.get(3).toString()), simInfo.getStep());
                        break;
                    case "dispenser":
                        map.addCellObject(position, new FBDispenserObject(simInfo.getBlockType(parameters.get(3).toString())), simInfo.getStep());
                        break;
                    case "entity":
                        String team = parameters.get(3).toString();
                        FBObjectType type;

                        if (!team.equals(agentInfo.getTeam())) { // not our team
                            type = FBObjectType.__FBEntity_Enemy;
                        } else if (position.equals(pov)) { // our team same position as agent => agent?
                            type = FBObjectType.__FBAgent;
                        } else {
                            type = FBObjectType.__FBEntity_Friend;
                        }
                        FBEntityObject entity = new FBEntityObject(type, team);
                        map.addCellObject(position, entity, simInfo.getStep());
                        if (type == FBObjectType.__FBAgent) {
                            entity.setName(agentInfo.getName());
                        }

                        break;
                    case "block":
                        map.addCellObject(position, new FBBlockObject(simInfo.getBlockType(parameters.get(3).toString()), attachedIndex.contains(relative)), simInfo.getStep());
                        break;
                    default:
                        HorseRider.yell(TAG, "processThings: " + agentInfo.getName() + " unprocessed: " + parameters);

                }
                /*HorseRider.inquire(TAG, "processThings: " + agentInfo.getName() + " " + String.format("thing: [%d,%d]: %s (%s) \n",
                        Integer.parseInt(parameters.get(0).toString()),
                        Integer.parseInt(parameters.get(1).toString()),
                        parameters.get(2), // what
                        parameters.get(3))); //team
                    */
            }
        }
    }

    public void setAgentInfo(AgentInfo agentInfo) {
        this.agentInfo = agentInfo;
    }

    public void setSimInfo(SimInfo simInfo) {
        this.simInfo = simInfo;
    }

}
